/*
  IDFIXES.CPP -- This file implements the fixing of vectors in the
  Interdyme framework.  Some functions here are members of Vector,
  others are members of FixBank. Note that FIXBANK.H has the full
  interface for the FixBank class.

  This file also contains an overload version of the function fix,
  a member of vector, which does the group fixes on imports.  The
  vector of domestic demand corresponding to the imports vector
  is passed as an argument, and this differentiates the vector
  group fix function from the normal fix function.

  In Dyme, version 2.2, we have added the capability to do matrix
  fixes.  We have borrowed from the Vector::fix routine to make
  the Matrix::fix.
*/

#include "dymesys.h"

#if defined(INCL_FIX)

const float ERRORR = -6.789e-7;

// Function prototypes, specific to import group fixes:
// return lambda as the value of the function.
float imnewton(float (*func)(float,float,short,float*,float*),
	float (*deriv)(float,short,float*,float*),
	float x1, float x2,
	float control, short n, float *m,
	float *flow, float toler=.001, float itermax=30);

short explanation(void);
short IsMissing(float);
short imsis(float control, short n, short *group, float *im,
	float *dd,float toler=.001);

float imfunc(float lambda,float control, short n, float *m, float *flow);
float imderiv(float lambda, short n, float *m, float *flow);

/*==========================================================================*/
/*      This is where the fixes are applied to the vectors.
	For fixes applying to a single sector:
	  c adds to fix to whatever is there
	  m multiplies fix by whatever is there
	  o, i, g, s all simply replace what is there by the fix.
	  e,f&h calculate the fix according to the expression, and then
		  replaces the sector item.
     n applies the fix relative to the base year value
	For fixes applying to a group
	  c sums up the group, adds the fix to the sum, and scales to
		 this total.
	  m multiplies all elements of the group by the fix
	  o, i, g, s all simply scale the sum of the elements to the fix.
	  e,f&h calculate the total control according to the follow expression,
		  and then scales the elements to sum to this result.
	Note: SectorFlag is an optional argument, set to 0 by default.
	  When it is 0, the fix routine is called normally, that is, with
	  both groups and single-sector fixes.  If it is > 0, then the
	  routine does only a fix for that single sector.  If it is equal to
	  -1, then only group fixes are done.
	Note: The return value from this routine is the last year of the fix,
	  if a single-sector fix was applied, otherwise it is -1.  If the
	  single-sector fix was of type 'c' or 'm' (add or mult), then
	  the option letter is returned instead.  This was written so that
	  a version of Seidel could implement different discrepancy years
	  for different sector.
		  In the case of no fixes being applied, this routine returns
	  a (-1).
	Changes were made to the vector fix functions with Interdyme 2.0,
	since "seq" is no longer the position within the Vam file, but
	rather the position of the vector within the mats[] array.

	Note, if a second argument is present, and it is a Vector, then
	this is interpreted to be an imports group fix function, which
	can be found immediately below this one.
*/
short Vector::fix(short t, short SectorFlag) {
char fixtype,mname[30],vname[30],*cp;
Chopper ch;    // for idrhs
short i,err,n,nin,macseq,sec,fixnumber,gn,*group,allgroup[MAXSEC];
short lvn,retval,start,end,fixbaseyr;
float FixVal,LagVal,FixBaseVal,SeriesBaseVal,total,MacVal,rhs,depend;
float *v1;

Vector Lagged(nelem);

/* The useall flag controls whether we can fix before lastdata. */
if(useall == 'y' && t <= LastDat) {
	return -1;
	}
/* We don't apply fixes when setrho=='y', as this section of the 
   model is *only* for calculating and updating errors for rho-adjustments.
*/
if(setrho == 'y')
   return -1;

if(fb.IsErr() ) 
	return(-2);


v1 = v - 1;  // v1 is a pointer to the elements of the vector such
			// v1[1] is the first element.

VamFile &vfp = *(mats[seq].vfp);
strcpy(vname,mats[seq].name);
lvn = vfp.findname(vname);
fixnumber = fb.FixPtr[lvn];
retval=-1;
if(fixesPlace == -1) {
	printf("If you are going to use fixes, you need to declare a vector\n");
	printf(" called fix in the VAM.CFG file, and a vector named fix must\n");
	printf(" be declared in MODEL.CPP!\n");
	return(-2);
	}

#ifdef INCL_EQFIX
float *f = new float[Interval+100];
char fnstring[500],*pname,name[500];
#endif

while(fixnumber >= 0 ){
	FixVal = mats[fixesPlace].vptr[fixnumber];
	fixtype = fb.fx[fixnumber]->command;
	start=fb.fx[fixnumber]->startfix;
	end=fb.fx[fixnumber]->endfix;
	if(t>=start && t<=end) {
		if(!fb.fx[fixnumber]->IsGroup){
			if(SectorFlag != -1) { // skip single-sector if == -1.
				// single-element fixes
				sec = fb.fx[fixnumber]->row;
				if(SectorFlag>0 && SectorFlag != sec) goto Next;
				retval=fixtype;  // returns option if not -1, +1, or lastact
				if(fixtype == 'm')
					v1[sec] *= FixVal;
				else if(fixtype == 'c')
					v1[sec] += FixVal;
            // Here to implement the "dynamic" fixes.  They are a little bit
            //   more complicated than the rest.
            else if(fixtype=='n') {   // dynamic index fix
               // Base year of fix is startfix
               if( (fixbaseyr=fb.Getstartfix(fixnumber)) >= t )
                  goto Next;
               err = fb.GetFixObs(fixnumber+1,fixbaseyr,FixBaseVal);
               if(IsMissing(FixBaseVal) || fabs(FixBaseVal) < 1.0e-7) {
                  cprintf("Error applying dynamic index fix on %s%d in %d: Base value of fix is zero!\r\n",
                     vname,sec,t);
                  return -1;
                  }
               err = vfp.GetVecElement(vname,sec,fixbaseyr,SeriesBaseVal);

               if(!IsMissing(FixVal)) {
                  depend = FixVal * SeriesBaseVal / FixBaseVal;
                  }
               v1[sec] = depend;
               }
            else if(fixtype=='w' || fixtype=='t') {  // dynamic growth rate fixes
               // For growth rates, base is 1 year before first fix.
               if( (fixbaseyr=fb.Getstartfix(fixnumber)-1) >= t )
                  continue;
               err = vfp.GetVecElement(vname,sec,t-1,SeriesBaseVal);
               depend = SeriesBaseVal * exp(FixVal/100.); // 2 -> .02
               v1[sec] = depend;
               }
				#ifdef INCL_EQFIX
				else if(fixtype == 'e') { // equation fix
					pname=name;
					MacFixNumber = -1;
					VecFixNumber = fixnumber;  // these are global
					cp=fb.GetFunction(fixnumber);
					if(cp!=NULL)
					strcpy(fnstring,cp);  
					strcat(fnstring,"\n");  // help chopper
					ch.SetSource(fnstring);
					err = ch.rhs(f,pname);
					rhs = f[Interval-1];
					v1[sec] = rhs;
					}
				else if(fixtype == 'f') { // follow fix
					pname=name;
					MacFixNumber = -1;
					VecFixNumber = fixnumber;
					cp=fb.GetFunction(fixnumber);
					if(cp!=NULL)
					strcpy(fnstring,cp);  
					strcat(fnstring,"\n");  // help chopper
					ch.SetSource(fnstring);
					err = ch.rhs(f,pname);

               // f is the value of the expression
               // FixVal is the extra growth rate to apply
               // LagVal is the lagged value of the variable
               //   (must not be zero or missing)

               // The hardest part of this fix for vectors is getting
               //   the lagged value.  

               //Lagged.load(vfp,vname,t-1);
               //LagVal = Lagged[sec];

               err = vfp.GetVecElement(vname,sec,t-1,LagVal);

					depend = LagVal*exp(FixVal/100+log(f[Interval-1]/f[Interval-2]));

               v1[sec] = depend;
					}
				else if(fixtype == 'h') { // share fix
					pname=name;
					MacFixNumber = -1;
					VecFixNumber = fixnumber;  // these are global
					cp=fb.GetFunction(fixnumber);
					if(cp!=NULL)
					strcpy(fnstring,cp);    
					strcat(fnstring,"\n");  // help chopper
					ch.SetSource(fnstring);
					err = ch.rhs(f,pname);
					rhs = f[Interval-1];
               if(IsMissing(rhs))
                  v1[sec] = MISSING;
               else
   					v1[sec] = FixVal * rhs;
					}
				#endif
				else {  // types 'o', 'g' or 's'
					v1[sec] = FixVal;
					retval=end;
               if(fb.fx[fixnumber]->IsOpt('f'))
                  retval = 'f';  // return value
					}
				} // end if == -1.
			} // end if group
		else{
			if( SectorFlag <= 0) {
				// A group
				gn = fb.fx[fixnumber]->groupnumber;
				// if it is the "all" group, generate the right sectors.
				if(strcmp(fb.fx[fixnumber]->groupname,"all")==0){
					for(i=0; i < nelem; i++)
						allgroup[i] = i+1;
					group = allgroup;
					n = nelem;
					}
				else{
					group = GroupShelter.gd[gn].GroupElems;
					n = GroupShelter.gd[gn].Nin;
					}
				if(fixtype == 'm'){
					for(i = 0;i < n; i++){
						v1[group[i]] *= FixVal;
						}
					}
				else if(fixtype == 'c'){
					total = 0;
					for(i = 0;i < n; i++)
						total += v1[group[i]];
					FixVal += total;
					rdscale(v1,n,group,FixVal);
					}
				else if(fixtype=='n' && !fb.fx[fixnumber]->IsOpt('s') ) {   // dynamic index fix
					// Base year of fix is startfix
					if( (fixbaseyr=fb.Getstartfix(fixnumber)) >= t )
						goto Next;
					err = fb.GetFixObs(fixnumber+1,fixbaseyr,FixBaseVal);
					if(IsMissing(FixBaseVal) || fabs(FixBaseVal) < 1.0e-7) {
						cprintf("Error applying dynamic index fix on %s%d in %d: Base value of fix is zero!\r\n",
							vname,sec,t);
						return -1;
						}
					// For group, SeriesBaseVal is sum in group
					Lagged.load(vfp,vname,fixbaseyr);
					SeriesBaseVal = 0.0;
					for(i=0;i<n;i++) 
						SeriesBaseVal += Lagged[group[i]];

					if(!IsMissing(FixVal)) {
						depend = FixVal * SeriesBaseVal / FixBaseVal;
						}
					rdscale(v1,n,group,depend);
					}              
				else if( (fixtype=='w' || fixtype=='t') && !fb.fx[fixnumber]->IsOpt('s') ) {  // dynamic growth rate fixes
					// For growth rates, base is 1 year before first fix.
					if( (fixbaseyr=fb.Getstartfix(fixnumber)-1) >= t )
						continue;
					// For group, SeriesBaseVal is sum in group
					Lagged.load(vfp,vname,t-1);
					SeriesBaseVal = 0.0;
					for(i=0;i<n;i++) 
						SeriesBaseVal += Lagged[group[i]];
					depend = SeriesBaseVal * exp(FixVal/100.); // 2 -> .02
					rdscale(v1,n,group,depend);
					}
				#ifdef INCL_EQFIX
				else if(fixtype == 'e') { // equation fix
					pname=name;
					MacFixNumber = -1;
					VecFixNumber = fixnumber;  // these are global
					cp=fb.GetFunction(fixnumber);
					if(cp!=NULL)
					strcpy(fnstring,cp);  
					strcat(fnstring,"\n");  // help chopper
					ch.SetSource(fnstring);
					err = ch.rhs(f,pname);
					rhs = f[Interval-1];
					rdscale(v1,n,group,rhs);
					}
				else if(fixtype == 'f' && !fb.fx[fixnumber]->IsOpt('s')) { // follow fix
					pname=name;
					MacFixNumber = -1;
					VecFixNumber = fixnumber;
					cp=fb.GetFunction(fixnumber);
					if(cp!=NULL)
					strcpy(fnstring,cp);  
					strcat(fnstring,"\n");  // help chopper
					ch.SetSource(fnstring);
					err = ch.rhs(f,pname);

					// f is the value of the expression
					// FixVal is the extra growth rate to apply
					// LagVal is the lagged value of the variable
					//   (must not be zero or missing)

					Lagged.load(vfp,vname,t-1);
					LagVal = 0.0;
					for(i=0;i<n;i++) 
						LagVal += Lagged[group[i]];

					depend = LagVal*exp(FixVal/100+log(f[Interval-1]/f[Interval-2]));
					v1[sec] = depend;
					rdscale(v1,n,group,depend);
					}
				else if(fixtype == 'h') { // share fix
					pname=name;
					MacFixNumber = -1;
					VecFixNumber = fixnumber;
					cp=fb.GetFunction(fixnumber);
					if(cp!=NULL)
					strcpy(fnstring,cp);  
					strcat(fnstring,"\n");  // help chopper
					ch.SetSource(fnstring);
					err = ch.rhs(f,pname);
					rhs = f[Interval-1];
					depend = FixVal * rhs;
					rdscale(v1,n,group,depend);
					}
				#endif
				else if(fb.fx[fixnumber]->IsOpt('s')) {  //"same" option, for 'g','s','i'
					// All items in group either grow the same, move the 
					//   same, or follow the same expression.
					// the way these are done, is same for "dynamic" or not
					if(fixtype=='g' || fixtype=='s' || fixtype=='w' || fixtype=='t') {
						// For growth rates, base is 1 year before first fix.
						if( (fixbaseyr=fb.Getstartfix(fixnumber)-1) >= t )
							continue;
						Lagged.load(vfp,vname,t-1);
						for(i=0;i<n;i++)  {
							SeriesBaseVal = Lagged[group[i]];
							depend = SeriesBaseVal * exp(FixVal/100.);
							v1[group[i]] = depend;
							}
						}
					else if(fixtype == 'i' || fixtype == 'n') {
						if( (fixbaseyr=fb.Getstartfix(fixnumber)) >= t )
							goto Next;
						err = fb.GetFixObs(fixnumber+1,fixbaseyr,FixBaseVal);
						if(IsMissing(FixBaseVal) || fabs(FixBaseVal) < 1.0e-7) {
							cprintf("Error applying dynamic index fix on %s%d in %d: Base value of fix is zero!\r\n",
								vname,sec,t);
							return -1;
							}
						Lagged.load(vfp,vname,fixbaseyr);
						for(i=0;i<n;i++) {
							SeriesBaseVal = Lagged[group[i]];
							if(!IsMissing(FixVal)) 
								depend = FixVal * SeriesBaseVal / FixBaseVal;
							v1[group[i]] = depend;
							}
						}
					else if(fixtype == 'f') {
   					pname=name;
   					MacFixNumber = -1;
   					VecFixNumber = fixnumber;
						cp=fb.GetFunction(fixnumber);
						if(cp!=NULL)
						strcpy(fnstring,cp);  
						strcat(fnstring,"\n");  // help chopper
						ch.SetSource(fnstring);
						err = ch.rhs(f,pname);

						Lagged.load(vfp,vname,t-1);
						for(i=0;i<n;i++)  {
							LagVal = Lagged[group[i]];
							depend = LagVal*exp(FixVal+log(f[Interval-1]/f[Interval-2]));
							v1[group[i]] = depend;
							}
						}
					} // end fixtype s
				else 
					rdscale(v1,n,group,FixVal);
				}
			} // end if SectorFlag <=0
		}
	Next:
	fixnumber = fb.fx[fixnumber]->Next;	
	}
cleanuptime:
#ifdef INCL_EQFIX
if(f)
	delete f;
#endif
err=err;
return retval;  // return last year of fix, -1, or 'c' or 'm'.
}
//-----------------------------------------------------------------------

short Vector::IsFixOpt(short sector, char opt) {

short fixnumber,lvn,start,end;
char fixtype,vname[30];

/* The useall flag controls whether we can fix before lastdata. */
if(useall == 'y' && t <= LastDat) {
	return -1;
	}
/* We don't apply fixes when setrho=='y', as this section of the 
   model is *only* for calculating and updating errors for rho-adjustments.
*/
if(setrho == 'y')
   return -1;

if(fb.IsErr() ) 
	return(-2);

VamFile &vfp = *(mats[seq].vfp);
strcpy(vname,mats[seq].name);
lvn = vfp.findname(vname);
fixnumber = fb.FixPtr[lvn];
if(fixesPlace == -1) {
	printf("If you are going to use fixes, you need to declare a vector\n");
	printf(" called fix in the VAM.CFG file, and a vector named fix must\n");
	printf(" be declared in MODEL.CPP!\n");
	return(-2);
	}

while(fixnumber >= 0 ){
   if(fb.fx[fixnumber]->row != sector)
      goto Next;
	start=fb.fx[fixnumber]->startfix;
	end=fb.fx[fixnumber]->endfix;
	if(t>=start && t<=end) {
		if(fb.fx[fixnumber]->IsOpt(opt))
      return 1;
      }
	Next:
	fixnumber = fb.fx[fixnumber]->Next;	
   }
return 0;
}
//=======================================================================
/* This is the Matrix version of fix */
short Matrix::fix(short t, char *mname) {

short lvn,err;
if(useall == 'y' && t <= LastDat)
	return(-1);
if(setrho == 'y')
   return(-1);

lvn = vf[0].findname(mname); // must be in vf[0]!
if(lvn<0)
	return -1;

err=DoFix(t,lvn,0);
return err;
}
//=======================================================================
short Matrix::fix(short t, short SectorFlag) {
short lvn,err;
if(useall == 'y' && t <= LastDat)
	return(-1);
if(setrho == 'y')
   return(-1);

VamFile *vfp = mats[seq].vfp;
lvn = vfp->findname(mats[seq].name);
if(lvn<0)
	return -1;
err=DoFix(t,lvn,SectorFlag);
return err;
}
//=======================================================================

short Matrix::DoFix(short t, short lvn, short SectorFlag) {

char fixtype,mname[20],vname[20];
short i,n,nin,macseq,row,col,fixnumber,gn,*group,allgroup[MAXSEC];
short retval,start,end;
float control,total,MacVal;
float *v1;

fixnumber = fb.FixPtr[lvn];
retval=-1;
if(fixesPlace == -1) {
	printf("If you are going to use fixes, you need to declare a vector\n");
	printf(" called fix in the VAM.CFG file, and a vector named fix must\n");
	printf(" be declared in MODEL.CPP!\n");
	return(-2);
	}

while(fixnumber >= 0 ){
	control = mats[fixesPlace].vptr[fixnumber];
	fixtype = fb.fx[fixnumber]->command;
	start=fb.fx[fixnumber]->startfix;
	end=fb.fx[fixnumber]->endfix;
	if( ! fb.fx[fixnumber]->IsMatrix )
		continue;
	if(t>=start && t<=end) {
		if(!fb.fx[fixnumber]->IsGroup){  // not a group
			if(SectorFlag != -1) { // skip single-sector if == -1.
				// single-element fixes
				row = fb.fx[fixnumber]->row;
				col = fb.fx[fixnumber]->col;
				if(SectorFlag>0 && SectorFlag != row) goto Next;
				retval=fixtype;  // returns option if not -1, +1, or lastact
				if(fixtype == 'm')
					m[row][col] *= control;
				else if(fixtype == 'c')
					m[row][col] += control;
				else {
					m[row][col] = control;
					retval=end;
					}
				} // end if == -1.
			} // end not a group
		}
	Next:
	fixnumber = fb.fx[fixnumber]->Next;
	}
return retval;  // return last year of fix, -1, or 'c' or 'm'.
}

//=======================================================================
/* This is the Pmatrix version of fix, new to version 2.35, 4/98
*/
short Pmatrix::fix(short t, char *mname) {

short lvn,err;
if(useall == 'y' && t <= LastDat)
	return(-1);
if(setrho == 'y')
   return(-1);
lvn = vf[0].findname(mname); // must be in vf[0]!
if(lvn<0)
	return -1;

err=DoFix(t,lvn,0);
return err;
}
//=======================================================================
short Pmatrix::fix(short t, short SectorFlag) {
short lvn,err;
if(useall == 'y' && t <= LastDat)
	return(-1);
if(setrho == 'y')
   return(-1);

VamFile *vfp = mats[seq].vfp;
lvn = vfp->findname(mats[seq].name);
if(lvn<0)
	return -1;
err=DoFix(t,lvn,SectorFlag);
return err;
}
//=======================================================================

short Pmatrix::DoFix(short t, short lvn, short SectorFlag) {

char fixtype,mname[20],vname[20];
short i,n,nin,macseq,irow,jcol,loccol,fixnumber,gn,*group,allgroup[MAXSEC];
short retval,start,end,legal;
float control,total,MacVal;
float *v1;

fixnumber = fb.FixPtr[lvn];
retval=-1;
if(fixesPlace == -1) {
	printf("If you are going to use fixes, you need to declare a vector\n");
	printf(" called fix in the VAM.CFG file, and a vector named fix must\n");
	printf(" be declared in MODEL.CPP!\n");
	return(-2);
	}

while(fixnumber >= 0 ){
	control = mats[fixesPlace].vptr[fixnumber];
	fixtype = fb.fx[fixnumber]->command;
	start=fb.fx[fixnumber]->startfix;
	end=fb.fx[fixnumber]->endfix;
	if( ! fb.fx[fixnumber]->IsMatrix )
		continue;
	if(t>=start && t<=end) {
		if(!fb.fx[fixnumber]->IsGroup){  // not a group
			if(SectorFlag != -1) { // skip single-sector if == -1.
				// single-element fixes
				irow = fb.fx[fixnumber]->row;
				jcol = fb.fx[fixnumber]->col;
				// Find location of this row and column in the Pmatrix,
				//  and make sure it is legal!
				if(rnc[irow]==0) {
					printf("Illegal pmatrix fix: %s(%d,%d)\n",fb.fx[fixnumber]->Vname,
						irow,jcol);
					return(-4);
					}
				legal = 0;
				for(i=0;i<rnc[irow];i++) {
					if(row[irow][i] == jcol) {
						legal = 1;
						loccol = i;
						break;
						}
					}
				if(!legal) {
					printf("Illegal pmatrix fix: %s(%d,%d)\n",fb.fx[fixnumber]->Vname,
						irow,jcol);
					return(-4);
					}

				if(SectorFlag>0 && SectorFlag != irow) 
					goto Next;
				retval=fixtype;  // returns option if not -1, +1, or lastact
				if(fixtype == 'm')
					p[irow][loccol] *= control;
				else if(fixtype == 'c')
					p[irow][loccol] += control;
				else {
					p[irow][loccol] = control;
					retval=end;
					}
				} // end if == -1.
			} // end not a group
		}
	Next:
	fixnumber = fb.fx[fixnumber]->Next;
	}
return retval;  // return last year of fix, -1, or 'c' or 'm'.
}

//=======================================================================
/* This is the version of fix that handles group fixes on imports.
	To differentiate this function from the normal fix function,
	the second argument passed is a Vector of domestic demands.
*/
short Vector::fix(short t, Vector& DomesticDemand) {
char fixtype,mname[20],vname[20];
short i,n,nin,macseq,sec,fixnumber,gn,*group,allgroup[MAXSEC];
short err,lvn,retval,start,end;
float control,total,MacVal;
float *dd = DomesticDemand.v; 

if(useall == 'y' && t <= LastDat)
	return(-1);
if(setrho == 'y')
   return(-1);
if(fb.IsErr() ) 
	return(-2);

VamFile *vfp = mats[seq].vfp;
lvn = vfp->findname(mats[seq].name);
fixnumber = fb.FixPtr[lvn];
retval=-1;
if(fixesPlace == -1) {
	printf("If you are going to use fixes, you need to declare a vector\n");
	printf(" called fix in the VAM.CFG file, and a vector named fix must\n");
	printf(" be declared in MODEL.CPP!\n");
	return(-2);
	}

while(fixnumber >= 0 ){
	control = mats[fixesPlace].vptr[fixnumber];
	fixtype = fb.fx[fixnumber]->command;
	start=fb.fx[fixnumber]->startfix;
	end=fb.fx[fixnumber]->endfix;
	if(t>=start && t<=end) {
		if(!fb.fx[fixnumber]->IsGroup){
			goto Next;
			}
		else{
			// A group
			gn = fb.fx[fixnumber]->groupnumber;
			// if it is the "all" group, generate the right sectors.
			if(strcmp(fb.fx[fixnumber]->groupname,"all")==0){
				for(i=0; i < nelem; i++)
					allgroup[i] = i+1;
				group = allgroup;
				n = nelem;
				}
			else{
				group = GroupShelter.gd[gn].GroupElems;
				n = GroupShelter.gd[gn].Nin;
				}
			if(fixtype == 'm'){
				// Control total to multiplied total, don't multiply each
				//   element.
				for(i = 0, total=0.;i < n; i++)
					total += v[group[i]-1];
				control *= total;
				err=imsis(control,n,group,v,dd);
				if(err == -1)
					goto Next;
				}
			else if(fixtype == 'c'){
				for(i = 0, total=0.;i < n; i++)
					total += v[group[i]-1];
				control += total;
				// Imsis: import scaling function.
				err=imsis(control,n,group,v,dd);
				if(err == -1)
					goto Next;
				}
			else {
				err = imsis(control,n,group,v,dd);
				if(err == -1)
					goto Next;
				}
			}
		}
	Next:
	fixnumber = fb.fx[fixnumber]->Next;
	}
return retval;  // return last year of fix, -1, or 'c' or 'm'.
}
//=======================================================================
//#define IMSISDEBUG
/* IMSIS:
	control is the control total, n is the number of elements in the
	group, group is the list of elements, im is the vector of imports,
	and dd is the vector of domestic demand.
*/

const float Top_Bracket=100.;

short imsis(float control, short n, short *group, float *im,
	float *dd,float toler) {

short j,err,loc,retval;
float lambda=1.,dif,f1,testsum;
float *m = new float[n];
float *flow = new float[n];
float (*pfn)(float,float,short,float*,float*);
float (*pder)(float,short,float*,float*);

pfn = imfunc;
pder = imderiv;

// Handle negative entries by first subtracting them from the control
//   total.  imnewton() knows not to include them in the scaling.

for(j=0;j<n;j++) {
	loc = group[j] - 1;
	flow[j] = dd[loc];
	if(im[loc]<0)
		control -= im[loc];
	m[j] = (flow[j]<1.0e-7)?0.:im[loc]/flow[j];
	}

// Try bracketing between 0. and Top_Bracket.
lambda=imnewton(pfn,pder,0.,Top_Bracket,control,n,m,flow,toler);
if (lambda == ERRORR)  {
	explanation();
	retval = -1;
	goto cleanup;
	}

// Calculate new import shares and revise imports.
//testsum=0.0;
for(j=0;j<n;j++) {
	if(m[j] > 0.)
		m[j] = m[j]/(m[j] + lambda*(1.-m[j]));
	loc = group[j]-1;
	im[loc] = m[j] * flow[j];
	//testsum += im[loc];
	}
retval = 0;
cleanup:

if(m)
	delete[] m;
if(flow)
	delete[] flow;

return retval;
}

//=====================================================================
// This small function evaluates the import share scaling function.
float imfunc(float lambda, float control, short n, float *m,
	float *flow) {
float f;
short j;
for(j=0,f=-control;j<n;j++) {
	if(m[j]<=0.)
		continue;
	if(fabs(flow[j])<1.0e-8)
		continue;
	f += m[j]/(m[j] + lambda*(1-m[j])) * flow[j];
	}
return f;
}
//=====================================================================
// Takes the derivative of the import share scaling function.
float imderiv(float lambda, short n, float *m, float *flow) {
float df;
short j;
for(j=0,df=0.;j<n;j++) {
	if(m[j]<0.)
		continue;
	df -= (m[j]*flow[j]*(1.-m[j])) /
		((m[j]+lambda*(1.-m[j]))*(m[j]+lambda*(1.-m[j])));
	}
return df;
}
//===============================================================
char * ExplainText=
" \n\n"
" There has been an error in imsis(), the function which controls\n"
" groups of sectors for imports.  This can be either because the\n"
" maximum number of iterations was exceeded in the newton algorithm\n"
" (rare) or getting the control to be satisfied would be impossible\n"
" without some import shares being outside the range (0,1).  The\n"
" latter is more common, and has occurred generally if the error\n"
" \"Root not bracketed in newton\" has appeared before this message.\n"
" Check your import group fix by comparing with domestic demands \n"
" for the corresponding industries.\n";
short explanation(void){
printf(ExplainText);
return 0;
}
//===============================================================
/* NEWTON -- The Newton-Raphson algorithm, the most efficient
	algorithm for finding a zero root in a function, when analytical
	derivatives are available.  The procedure is to take the tangent
	to the function, at the iteration point, follow this tangent
	to the x-axis, and then re-evaluate the function at this new
	point.  Now take the tangent at the new point, and the procedure
	iterates until convergence.

	This version of the algorithm, culled from "Numerical Recipes",
	requires the user to supply the function and the derivative.
	Like secant(), the function starts with two bracket points,
	x1 and x2.  This "fail-safe" version takes a bisection step
	whenever Newton-Raphson would take the solution out of bounds
	or whenever Newton Raphson is not reducing the size of the brackets
	rapidly enough.

	This version also skips elements of the imports vector which are
	negative.  It assumes that the control has already been adjusted 
	in the calling function.  See details in imfunc() and imderiv().
*/

// func and deriv are function and derivative evaluators.

float imnewton(float (*func)(float,float,short,float*,float*),
					float (*deriv)(float,short,float*,float*),
					float x1, float x2,
					float control, short n, float *m,
					float *flow, float toler, float itermax) {

float xl, xh, fl, fh, f, df, dx, dxold, swap, rts, temp;
short iter;

// Pick the bound with the smallest absolute function value as the
//   "recent guess".
fl = (*func)(x1,control,n,m,flow);
fh = (*func)(x2,control,n,m,flow);
if(fl*fh > 0.) {
	printf("Root is not bracketed in newton!\n");
	return(ERRORR);
	}
// Orient the search so that f(xl) < 0.
if(fl < 0.0) {
	xl = x1;
	xh = x2;
	}
else {
	xh = x1;
	xl = x2;
	swap=fl;
	fl=fh;
	fh=swap;
	}
// Initialize the guess for the root, the "step-size before last"
//  and the last step.

rts = 0.5 * (x1 + x2);  // bisection
dxold = fabs(x2-x1);
dx = dxold;
f =  (*func)(rts,control,n,m,flow);
df = (*deriv)(rts,n,m,flow);

for(iter=1;iter<itermax;iter++) {
	// Bisect if Newton is out of range, or not decreasing fast enough.
	if((((rts-xh)*df-f)*((rts-xl)*df-f) >= 0.0)
				|| (fabs(2.0*f) > fabs(dxold*df))) {
		dxold=dx;
		dx=0.5*(xh-xl);
		rts=xl+dx;
		if(xl==rts)     // Change in root is neglible.
			return rts;
		}
	else {  // Newton step acceptable, take it.
		dxold=dx;
		dx = f/df;
		temp = rts;
		rts -= dx;
		if(temp==rts)
			return rts;
		}
	// Convergence criterion:
	if(fabs(dx) < toler)
		return rts;
	// New function and derivative evaluation for this iteration.
	f = (*func)(rts,control,n,m,flow);
	df = (*deriv)(rts,n,m,flow);
	#ifdef IMSISDEBUG
	cprintf("Iter: %d   x = %6.4f  f(x) = %6.4f  df = %6.2f.\r\n",iter,rts,f,df);
	#endif
	// Maintain the bracket on the root.
	if(f < 0.0) {
		xl = rts;
		fl = f;
		}
	else {
		xh = rts;
		fh = f;
		}
	} // end iter;
printf("Maximum number of iterations (%d) exceeded in newton().\n",
	itermax);
return ERRORR;
}
//=======================================================================
/*      GetFixType - This method of vector determines if a fix has been
	specified for the sector number which is passed as an argument.
	If so, it returns the fix type ('o','g',s','i','c',m'), otherwise
	it returns a 0.
*/
char Vector::GetFixType(short Sector) {
char fixtype;
short sec,fixnumber,lvn;
char FixType=0;

VamFile *vfp = mats[seq].vfp;
lvn = vfp->findname(mats[seq].name);
fixnumber = fb.FixPtr[lvn];

while(fixnumber >= 0 ){
	fixtype = fb.fx[fixnumber]->command;
	if(!fb.fx[fixnumber]->IsGroup){
		// single-element fixes
		sec = fb.fx[fixnumber]->row;
	if(Sector>0 && Sector != sec) 
		goto Next;
	FixType=fixtype;  // We will return the last fix type of this sector.
	} 
	Next:
	fixnumber = fb.fx[fixnumber]->Next;
	}
return FixType; 
}
//===================================================================
short FixBank::GetFixObs(short fixno, short t, float& value){
	long posit,temp,bnkheadsize;
	char cnpy;
	short base,err;

	fvfp->GetVecElement("fix",fixno,t,value);

	return 0;
}
/*==========================================================================*/
short FixBank::Getstartfix(short fixno) {
if(fixno < 0 || fixno > nseries)
	return -1;

return fx[fixno]->startfix;
}
/* ========================================================================= */
short FixBank::PutInGroupNums(void) {
short i,j,n,found,err = 0;
char search[25];

n = GroupShelter.GetNGroups();
for(i = 0; i < nseries; i++){
	found=FALSE;
	if(!fx[i]->IsGroup){
		fx[i]->groupnumber = -1;
		continue;
		}
	strcpy(search,fx[i]->groupname);
	for(j = 0; j < n; j++){
		if(strcmp(search,GroupShelter.gd[j].GroupTag) == 0){
			fx[i]->groupnumber = j;
			found=TRUE;
			break;
			}
		}
	if(!found) {
		printf("Cannot find group %s.\n",search);
		err -=1;
		}
	}
return(err);
}
/*==========================================================================*/
Fix::Fix() {
short i;
Next=-1;
IsGroup=0;
groupnumber=0;
row=0;
col=0;
IsMatrix=0;
MultiFix=0;
strcpy(groupname,"");
strcpy(Vname,"");
for(i=0;i<15;i++)
	Flags[i] = 0;
startfix=-1;
endfix=-1;
funlen=0;
funpos=0;
function=0;
}
/*==========================================================================*/
Fix::~Fix() {
if(function)
	delete function;
}
/*==========================================================================*/
Fix::Fix(Fix& f) {  // typical copy constructor
Next=f.Next;
IsGroup=f.IsGroup;
IsMatrix=f.IsMatrix;
//MultiFix=f.MultiFix;
row=f.row;
col=f.col;
strcpy(groupname,f.groupname);
groupnumber=f.groupnumber;
command=f.command;
strcpy(Vname,f.Vname);
strcpy(Flags,f.Flags);
startfix=f.startfix;
endfix=f.endfix;
funlen=f.funlen;
funpos=f.funpos;
if(funlen>0) {
	function=new char[funlen];
	strcpy(function,f.function);
	}
}
/*==========================================================================*/
Fix& Fix::operator=(Fix& f) {  // typical operator =
Next=f.Next;
IsGroup=f.IsGroup;
IsMatrix=f.IsMatrix;
//MultiFix=f.MultiFix;

row=f.row;
col=f.col;
strcpy(groupname,f.groupname);
groupnumber=f.groupnumber;
command=f.command;
strcpy(Vname,f.Vname);
strcpy(Flags,f.Flags);
startfix=f.startfix;
endfix=f.endfix;
funlen=f.funlen;
funpos=f.funpos;
if(funlen>0) {
	function=new char[funlen];
	strcpy(function,f.function);
	}
return(*this);
}
/*==========================================================================*/
short Fix::ReadIn(FILE *ip) {
	short err=0;
	err+=fread(&Next,sizeof(short),1,ip);
	err+=fread(&IsGroup,sizeof(short),1,ip);
	err+=fread(&IsMatrix,sizeof(short),1,ip);
	err+=fread(&MultiFix,sizeof(short),1,ip);
	err+=fread(&row,sizeof(short),1,ip);
	err+=fread(&col,sizeof(short),1,ip);	
	err+=fread(groupname,sizeof(char),25,ip);
	err+=fread(&groupnumber,sizeof(short),1,ip);
	err+=fread(&command,sizeof(char),1,ip);
	err+=fread(Vname,sizeof(char),15,ip);
	err+=fread(Flags,sizeof(char),15,ip);
	err+=fread(&startfix,sizeof(short),1,ip);
	err+=fread(&endfix,sizeof(short),1,ip);
	err+=fread(&funlen,sizeof(short),1,ip);
	err+=fread(&funpos,sizeof(long),1,ip);
//   err+=fread(&function, sizeof(char *),1,ip);
//   err+=fread(this,sizeof(Fix),1,ip);
	return err;
}
/*==========================================================================*/
short Fix::IsOpt(char c) {  // returns true if option is present
char *cp;
for(cp=Flags;*cp;cp++) {
	if(*cp==c)
		return 1;
	}
return 0;
}
/*==========================================================================*/
// Opens an existing FixBank in the current Dyme environment, where VamFile
//    is not used to talk to DYME.BIN.

/*==========================================================================*/
FixBank::FixBank(){
short i;
for(i=0;i<MAXFIX;i++)
	fx[i]=NULL;
FixPtr=NULL;
FixMat=NULL;
}
/*==========================================================================*/
void FixBank::Load(char *rootname, VamFile& vf){

short len,i,err;
long posit,cumpos,funcspos;
char fname[120];
size_t rcount=0, siz=0;

//printf("Constructing object for FixBank %s.\n",rootname);

FixPtr=NULL;
FixMat=NULL;
indname=NULL;
bnktitle=NULL;
bnkname=NULL;
opened=-1;

strcpy(fname,"FixBank");
classname=new char[strlen(fname)+1];
strcpy(classname,fname);
/* Open .fin file and initialize. */
strcpy(fname,rootname);
strcat(fname,".fin");
bnkname= new char[strlen(fname)+1];
strcpy(bnkname,fname);

nvecs = vf.Getnvecs();
jahra = (unsigned short)vf.Getstartdate();
jahra = (jahra<1900)?jahra+1900:jahra;
nopy = 1;
nobs = (unsigned short)(vf.Getenddate()-vf.Getstartdate())+1;

if((ipbank = fopen(bnkname,"rb")) == 0) {
	printf(msgbk2,bnkname);
	goto failure;
	}
posit = 0;
if((fseek(ipbank,posit,0)) != 0) {   /* posit at beginning of file */
	printf(msgbk20);
	goto failure;
	}

if((rcount=fread(&nseries,sizeof(short),1,ipbank)) == 0) {
	printf(msgbk20);
	goto failure;
	}
if((rcount=fread(&nvecs,sizeof(short),1,ipbank)) == 0) {
	printf(msgbk20);
	goto failure;
	}
FixPtr=new short[nvecs];
if((rcount=fread(FixPtr,sizeof(short),nvecs,ipbank)) == 0) {
	printf(msgbk20);
	goto failure;
	}

// allocate fixes array
for(i=0;i<MAXFIX;i++)
	fx[i]=new Fix;

if(nseries == 0) goto success;
for(i=0;i<nseries;i++) {
	if((err=fx[i]->ReadIn(ipbank))== 0) {
		printf(msgbk20);
		goto failure;
		}
	}
nfunc=0;   cumpos=0;  
funcspos = (nvecs+2)*sizeof(short) + sizeof(Fix)*nseries;
for(i=0;i<nseries;i++) {
	if(fx[i]->funlen>0) {
		nfunc++;
		fx[i]->funpos = cumpos;
		cumpos += fx[i]->funlen;  // = strlen(func) + 1
		posit = funcspos + fx[i]->funpos;
		if((fseek(ipbank,posit,0)) != 0) {
			printf(MSGBK(15));
			goto failure;
			}
		fx[i]->function = new char[fx[i]->funlen];
		fread(fx[i]->function,fx[i]->funlen,1,ipbank);
		}
	}

fvfp = &vf;   // store VamFile pointer for some other operations.
banktype = 'f';  siz=siz; rcount=rcount; err=err; // prevent dumb warnings
goto success;
failure:
	opened = -1;
	dymetap();
	return ;
success:
	opened = 1;
	return;
}

/*==========================================================================*/
// Destructor: free up memory allocated in constructor and close data files.
FixBank::~FixBank() {

short i;
//printf("Destructor for FixBank called.  Freeing up memory.\n");
if(classname) 
	delete[] classname;
if(bnkname) 
	delete[] bnkname;
if(bnktitle) 
	delete[] bnktitle;
if(indname) 
	delete[] indname;
for(i=0;i<MAXFIX;i++)
	if(fx[i]) 
		delete fx[i];
if(FixPtr) 
	delete[] FixPtr;
if(FixMat) 
	delete FixMat;
if(MsgObject)
	delete MsgObject;
fclose(pbank);
fclose(ipbank);
}
/*==========================================================================*/
char *FixBank::Setbnktitle(const char *s) {

short err;
bnkHeader Hdr;

bnktitle = new char[strlen(s)+1];
strcpy(bnktitle,s);
Hdr.jahra=jahra;
Hdr.nobs=nobs;
Hdr.stper=stper;
strcpy(Hdr.bnktitle,bnktitle);
// update bank with title, too.
if((fseek(pbank,0L,0)) != 0) {   /* posit at begin of file */
	printf(msgbk15);
	return(NULL);
	}
if((err=fwrite( (void*)&Hdr,sizeof(Hdr),1,pbank))==  0){
	printf(msgbk1,bnkname);
	return(NULL);
	}
err=err;
return bnktitle;
}
/*==========================================================================*/
char* FixBank::GetFunction(short fixno) {
char *cp;
if(fx[fixno]->funlen==0)
	return NULL;
return (cp=fx[fixno]->function);
}

#endif  // end fixes section
