//  MACFwXBNK.H -- Class declarations for MacFixBank objects.
//    Fix types are as follows:
//       'k' - skip(sk), uses macro data already in G bank
//       'o' - override(ovr), like LIFT AK fix.
//       'm' - multiply(mul), like LIFT MK fix.
//       'c' - constant(cta), like LIFT CK fix.
//       'i' - index(ind), like LIFT IK fix.
//       'g' - growth(gro), like LIFT RK fix.
//       'p' - percgrowth(pgro), growth rate with simple percent change.
//       's' - stepgrowth(stp), like LIFT GK fix.
//       'r' - rho adjustment, like in G/Build
//       'e' - equation fix
//       'f' - follow fix
//       'h' - share fix
//       'n' - dynamic index fix
//       'w' - dynamic growth rate fix
//       't' - dynamic step growth fix
//
//    The results of the MACFIXER program will go into two files:
//
//    The index file, .ind, will contain nseries (the number of fixes)
//    at the front of the file, followed by a list of names, followed
//    by as many nseries MacFix records.
//
//    The macfixbank file, .bnk, which contains the fixes in time series
//    format, arranged much as in a simple G bank.
//
//    The key to the model operation is the modify function, which has
//    now been modified to handle all the above types of fixes.
//
//    The MACFIXER program also needs to open up a companion G databank,
//    which is presumably the dyme workspace bank that will be used when
//    the model is running.  Index and growth fixes will be calculated
//    using the data in this Gbank.


#if (!defined(MACFXBNK_H))
#define MACFXBNK_H 1

class DataBank;
class GBank;
class Chopper;


// missing value code
#define MISSING    (-.0000010)

/* The MacFix class is used in the MacFixer program to read and store
   the Macrofixes.
*/
struct MacFix {
   char command;   /* 'k','o','m','c','i','g','s','p','e','f', */
                   /*     'h','r','n','w' or 't'(see above) */
	char Mname[15]; // Macro variable name
	short startfix;   // starting year of fix - subtract jahra to get index.
	short endfix;     // ending year of fix
	short funlen;     // function length
	long funpos;      // function position from start of functions
	char *function;
	char *IAm(char *fxstring);
	char *ForShort(char *sname);
	MacFix();
	MacFix(MacFix&); // copy constructor
	MacFix& operator=(const MacFix&);
	~MacFix();
	short ReadIn(FILE* fp);
	short WriteOut(FILE *fp);
};

/* The MacFixBankis the bject to handle transactions with Dyme 
   macrovariable fix file.
*/

class MacFixBank : public DataBank    {
protected:
	MacFix *mfx;  // for setting up array of fixes, usually MAXFIX, or nseries.
	// The following have been added to make this more like a G databank.
	unsigned short ncin;
	char *names;
	unsigned short *indx;
	short nfunc; // number of functions in file
public:
	enum {MAXMACFIX=500, NOMAX=650, NCMAX=27500, NSMAX=2500};
	// constructors
	MacFixBank(void); // a do-nothing constructor for making the thing global
	MacFixBank(const char *rootname,const char *status,GBank& gb);
	MacFixBank(const char *rootname,GBank& gb);
	MacFixBank(MacFixBank& mf);// copy constructor
	short Load(const char *rootname) ;
	char *Setbnktitle(const char *s);
	// destructor
	~MacFixBank();

	// index I-O
	short windex(void);    // write and read index of MacFix
	short rindex(void);
	short FSearch(MacFix& f);

	// Put and Get
	short AddFix(MacFix& f, float *ss,char wrindex_flag='y');
	short GetFix(short which,float *ss);
	short getser(char *nam,float *f,short SeriesLength=-1){return 0;}

	short PutFix(short which,float *ss);
	short PrintFix(FILE *chk, MacFix& f, float *ss);
	short StoreFixes(void);
	// The following calculate fixes as actual, moving forward actual points
	//    in the MacFixBank file.
	short GrowthFix(MacFix& fx,float *ss,GBank& gb,FILE *chk);
	short IndexFix(MacFix& fx,float *ss, GBank& gb,FILE *chk);
	short RhoFix(MacFix& fx, float rho);

	short ReadFix(FILE *iin, FILE *chk, short& styr, short& nper, float ss[],
				MacFix& f);
	short ReadFix(Chopper& ch, FILE *chk, short& styr, short& nper, float ss[],
				MacFix& f);
	short GetFixName(short i, char* name);  // returns "fixname", given index
	short GetRhoFix(short which, float& rho, short& rhostart);
	short GetFixObs(short which, short t, float& value);
	short Announce(short bankno=0);
	short ListNames(FILE*,char*,short wide=1){return 0;}
	char* GetFunction(short fixno);
   short Getstartfix(short fixno);
	
	// inline functions
	short Getnobs(void) {return nobs;}
	short Getnfix(void) {return nseries;};
};

///////////////////////////////////////////////////////////////////////////
// ***** Now, the classes for using macro fixes within the model.
class MacroFix {
  public:
	char *MacroName;  // Name of MacroVariable
	short FixNumber;    // order of the fix in the MacFixBank.
	short MacroNumber;  // "seq" number, used in Tseries object.
	char FixType;     // Type of fix ("komcigsrefhnwt")
	short RhoStart;     // Starting year of rho-adjustment
	float Rho;        // Value of rho, if any.
	float Error;      // Miss of eq., then adjustment to be added.
	float FixValue;    // Value of fix to be applied, if any.
							//  (This is updated every year).
	short funlen;     // length of function (=0 if none)
	long funpos;      // position of function at end of index file
	char *function;
	// Default constructor:
	MacroFix(char *MName);
	MacroFix(MacroFix& mf);
	~MacroFix();
	};

class MacroFixHome {
  public:
	short nfix, nmacro;  // number of fixes, number of macrovariables in model
	MacroFix *MFixVec[MacFixBank::MAXMACFIX]; // Array of MacroFixes.

	// Default constructor:
	MacroFixHome();
	MacroFixHome(MacFixBank& mfb, short ns);
	MacroFixHome(MacroFixHome& mfh);
	short Load(MacFixBank& mfb, GBank& gb);
	~MacroFixHome();
	short FixUpdate(short t, MacFixBank& mfb);  // Updates fixes for given year.
	short GetRho(short seq, short& fixnum, float& rho, short& rhostart, float& error);
	short PutError(short fixnum, float error);
	short ApplyFixes(short seq, float& depend);
	short IsFixed(short seq);
	};

#endif // end MACFXBNK_H class definitions

