/* EQUATION.CPP -- This file and EQUATION.H implement the Equation class,
	for detached-coefficient equations in Interdyme.
*/

#include "dymesys.h"  // Include Interdyme System files

// This entire class may be #ifdef'd out if you do not #define INCL_EQUATION

#if defined (INCL_EQUATION)
// =========================================================================
//      Equation methods:
// =========================================================================

/* Equation() -- read a group of equations from a file.  First line of file
	is: neq ncol rhostart
	each equation is then as produced by ipch command in G, namely
		label sector type niv
		psn1 ... psnniv
		rho coef1 ... coefniv
	The optional "yr' is used to update rhostart.
	*/
Equation::Equation(char *fname, int yr)
{
    sectors = 0;
    r(fname,yr);
    }
int Equation::r(char *fname,int yr)
{
	char card[241],*s,typ;
	short i,j,psn[40],sector,niv;
	FILE *fp;

	sectors=0;
	filename=0;
	types=0;
	rhos=0;
	errors=0;
	coefs=0;
	sectors=0;

	filename = new char[strlen(fname)+1];
	strcpy(filename,fname);
	if((fp = fopen(filename,"r")) == 0){
		printf("Cannot open %s.\n",filename);
		return(-1);
		}
	fgets(card,241,fp);
	i = 0;
	getint(neq,fp,card,i);
	getint(ncol,fp,card,i);
	getint(rhostart,fp,card,i);
	rhostart=(yr>rhostart)?yr:rhostart;
//	rhostart=(yr>0)?yr:rhostart;
	sectors = ivector(1,neq);
	rhos = vector(1,neq);
	errors = vector(1,neq);
	types = new char[neq+1];
	coefs = matrix(1,neq,1,ncol);
	// zero out
	for(i = 1; i <= neq; i++){
		sectors[i] = 0;
		rhos[i] = 0;
		errors[i] = 0;
		types[i] = ' ';
		for(j = 1; j <= ncol; j++)
			coefs[i][j] = 0;
		}
	// read individual equations
	eqcount = 0;
	while(fgets(card,241,fp) != NULL){
		if(strlen(card) <= 1) break;
		eqcount++;
		if(eqcount > neq){
			printf("%d is too many equations in %s.\n",eqcount,filename);
			exit(1);
			}
//		s = strchr(card,' ');
//Values can not be scan in sector. Others are OK. Why?
//		sscanf(s," %d %c %d",&sector,&typ,&niv);
//		sectors[eqcount]=sector;
//This works, with sector replaced with sectors[eqcount]
		s = strchr(card,' ');
//Values can be correctly scanned into sectors[eqcount] and others.
		sscanf(s," %d %c %d",&sectors[eqcount],&typ,&niv);
		sector=sectors[eqcount];
		if(niv > ncol){
			printf("Sector %d has too many coefficients in %s.\n",sector,filename);
			exit(1);
			}
		types[eqcount] = typ;
		fgets(card,241,fp);
		j = 0;
		for(i = 0; i < niv; i++){
			getint(psn[i],fp,card,j);
			if(psn[i] < 1 || psn[i] > ncol){
				printf("Equation for sector %d in %s puts a coefficient in column %d,\n",
					sector,filename,psn[i]);
				printf("which is outside the coefficient matrix.\n");
				}
			}
		fgets(card,241,fp);
		j = 0;
		getfloat(rhos[eqcount],fp,card,j);
		for(i = 0; i < niv; i++)
			getfloat(coefs[eqcount][psn[i]],fp,card,j);
		}
	fclose(fp);
	}
//======================================================================
// Equation destructor
Equation::~Equation() {
	#ifdef DBG_DESTRUCT
	printf(" Calling destructor for Equation.  \r\n"
			"filename = %s  \r\n", filename);
	#endif
	if(sectors)
		free_ivector(sectors,1);
	if(rhos)
		free_vector(rhos,1);
	if(errors)
		free_vector(errors,1);
	if(types)
		delete[] types;
	if(filename)
		delete[] filename;
	if(coefs)
		free_matrix(coefs,1,neq,1);
}
//======================================================================
// Copy constructor for Equation --> prevents bugs caused by passing
//   by value, not by reference. 12/27/94
Equation::Equation(Equation& e){
	short i,j;

	neq=e.neq;
	ncol=e.ncol;
	rhostart=e.rhostart;
	eqcount=e.eqcount;

	sectors = ivector(1,neq);
	rhos = vector(1,neq);
	errors = vector(1,neq);

	for(i=1;i<=neq;i++) {
		sectors[i]=e.sectors[i];
		rhos[i]=e.rhos[i];
		errors[i]=e.errors[i];
		}
	types = new char[neq+1];
	for(i=0;i<=neq+1;i++)
		types[i]=e.types[i];
	coefs = matrix(1,neq,1,ncol);
	for(i=1;i<=neq;i++)
		for(j=1;j<=ncol;j++)
			coefs[i][j]=e.coefs[i][j];
}
//======================================================================
//Default constructor
Equation::Equation()
{
	neq = 0;
	ncol = 0;
	errors = 0;
	sectors = 0;
	rhos = 0;
	types = 0;
	coefs = 0;
	filename=0;
	}
//======================================================================
// SetConstantTerm: Override estimated value of rho to get constant 
//		term adjustment.
void Equation::SetConstantTerm(void)  {
if(rhos) 
	for(short i=1;i<=neq;i++) 
		rhos[i]=1.0;
}
//======================================================================
// ZeroRho: Override estimated value of rho and make it zero.
void Equation::ZeroRho(void)  {
if(rhos) 
	for(short i=1;i<=neq;i++) 
		rhos[i]=0.0;
}
//======================================================================
// SetRho: Override estimated value of rho and make it specified value.
void Equation::SetRho(float value)  {
if(rhos) 
	for(short i=1;i<=neq;i++) 
		rhos[i]=value;
}
//======================================================================
char Equation::type(short i)
{
	return types[i];
	}
//======================================================================
float Equation::rho(short i)
{
	return rhos[i];
	}
//======================================================================
short Equation::sec(short i)
{
	return sectors[i];
	}
//======================================================================
short Equation::GetLoc(short sec) {
short i,iwhich;
iwhich= -1;
for(i=1;i<=eqcount;i++) {
	if(sectors[i]==sec) {
		iwhich=i;
		break;
		}
	}
return(iwhich);
}
//======================================================================
/* rhoadj() -- perform rho adjustment on predicted values from equation.*/
float Equation::rhoadj(float value, float xi, short i,float rhoval)
{

	/* value is what was computed by the equation
		xi is what was already in the vector
		i is the element number being worked on.
		the return value is the rho-adjusted forecast.
		*/

	float ret;

	if (useall == 'y' && t < rhostart)
		return(xi);
	if((useall == 'y' && t == rhostart) || useall == 'n' && t == godate){
		if(setrho == 'y')
            // Set errors[i] ready to be simply added to the predicted value in
            // the next period.
			errors[i] = (xi - value)*(fabs(rhoval)>1?rhos[i]:rhoval);
		return(xi);
		}
	else{  // Make a rho adjustment and update the error term.
		if(setrho == 'y') {
			/* When setrho == 'y', value should not change from xi */
			errors[i] = (fabs(rhoval)>1?rhos[i]:rhoval)*errors[i];
			// try to get rid of floating point underflow..
			if(fabs(errors[i])<1.0e-8)
				errors[i]=0.0;
			ret = xi;
			}
		else  {
			ret = value + errors[i];  // default
			}
		return(ret);
		}
	}
//======================================================================
float * Equation::operator[](short i)
{
	if(i < 1 || i > neq){
		printf("Illegal equation number %d. Returning first equation.\n");
		return coefs[1];
		}
	return(coefs[i]);
	}

#endif  // end equations section
