//  FIXBANK.H -- Class declarations for FixBank objects.
//    Fix types are as follows:
//       'o' - override(ovr), like LIFT AK fix.
//       'm' - multiply(mul), like LIFT MK fix.
//       'c' - constant(cta), like LIFT CK fix.
//       'i' - index(ind), like LIFT IK fix.
//       'g' - growth(gro), like LIFT RK fix.
//       's' - stepgrowth(stp), like LIFT GK fix.
//       'l' - elementgrowth(lgr), growth rate fix for each element of a group grow
//       'e' - equation fix, better than LIFT E fix.
//       'f' - follow(fol) fix
//       'h' - share(shr) fix
//       'n' - dynamic index fix
//       'w' - dynamic growth rate fix
//       't' - dynamic step growth fix
//
//    The results of the FIXER program will go into two files:
//
//    The index file, .fin, will contain nseries (the number of fixes)
//    at the front of the file, followed simply by as many nseries Fix
//    records.
//
//    The fixes themselves will go into the VAM.BIN file, in a special
//    vector, called "fix".  This vector will be declared in the model
//    also, and loaded each year like any other vector.
//
//    The key to the model operation is the functions that apply the
//       fixes.  These could be members of Vector class.

#if (!defined(FIXBANK_H))
#define FIXBANK_H 1

#ifdef __DMPI16__
#define MAXFIX  2000
#else
#define MAXFIX  500
#endif

#include "databank.h"
#include "vamfile.h"
#include "groups.h"

#if !defined(BNKHEADER)
#define BNKHEADER 1
struct bnkHeader {
   short jahra, stper, nobs;
   char bnktitle[80];
   };
#endif
struct Fix {
   short Next;       // Index in array of next fix with this vname.
   short IsGroup;
   short IsMatrix,MultiFix;  // Multifix becomes many fixes!
   short row,col;  // We have either Group Or Sector for row and col
   char groupname[25];
   short groupnumber; //for use by dyme
   char command;   // 'o','m','c','i','g','s','e','f' or 'h' (see above)
   char Vname[15]; // Vector name;
   char Flags[15]; // Flag string
   short startfix;   // starting year of fix - subtract jahra to get index.
   short endfix;     // ending year of fix
   short funlen;     // function length
   long funpos;      // function position from start of functions
   char *function;

   char *IAm(char *fxstring);
   char *ForShort(char *sname);
   Fix(); //constructor
   ~Fix();
   Fix(Fix&);  // copy constructor
   Fix& operator=(Fix&);  // operator =
   short ReadIn(FILE *ip);
   short IsOpt(char c);
};

class FixBank : public DataBank    {
// FixBank: Object to handle transactions with Dyme vector fix file.

protected:
   Fix *fx[MAXFIX];     // for setting up array of fixes, usually MAXFIX, or nseries.
   short nvecs;   // number of vectors in vam.  length of FixPtr vector
   short *FixPtr; /* will be allocated size nvecs, and hold the index of the
                     first fix of each lvn number ("am"=1).  If there
                     are no fixes for the corresponding item, this number
                     is (-1).   */
   Matrix *FixMat;
   VamFile *fvfp;  // pointer to Vam file holding fixes 
   short nfxrows, nfxcols;
   short nfunc;
public:
   // constructors
   FixBank(void); // a do-nothing constructor for making the thing global
   FixBank(const char *rootname,const char *status,VamFile& vf);
   FixBank(const char *rootname,VamFile& vf);
   short PutInGroupNums(void);
   void Load(char *rootname, VamFileDesc vfd);  // This initializes fixes from the .fin file.
   void Load(char *rootname, VamFile& vf);     // Newer alternative.
   virtual char *Setbnktitle(const char *s);
   // destructor
   ~FixBank();

   // index I-O
   short windex(void);    // write and read index of Vnames, GroupOrSec,
   short rindex(void);    //    and command.
   short FSearch(Fix& f);
   unsigned short GetBaseYr(void) {return jahra;}

   // Put and Get
   short AddFix(Fix& f, float *ss,VamFile& vf,char wrindex_flag='y');
   short GetFix(short which,float *ss);
   short PutFix(short which,float *ss);
   short PrintFix(FILE *chk, Fix& f, float *ss);
   short StoreFixes(VamFile &vf);
   // The following calculate fixes as actual, moving forward actual points
   //    in the VAM file.
   short GrowthFix(Fix& fx,float *ss,VamFile& vf,GroupHome& groups);
   short IndexFix(Fix& fx,float *ss,VamFile& vf,GroupHome& groups);

   short ReadFix(Chopper& ch,VamFile& vam, short& styr, short& nper, float ss[],
            Fix& f);

   short getser(char *nam, float *f, short SeriesLength) {return 0;};
   short ListNames(FILE*, char*, short wide) {return 0;};

   short SetNext(VamFile& vf);
   short WrtTTL(char *ttlfile);

   // inline functions
   short Getnobs(void) {return nobs;}
   short Getnfix(void) {return nseries;};
   char* GetFunction(short fixno);
   short Getstartfix(short fixno);
   short GetFixObs(short fixno, short t, float& fixval);

   friend short Vector::fix(short t, short SectorFlag=0);
   friend short Vector::fix(short t, Vector& DomesticDemand);
   friend short Matrix::DoFix(short t, short lvn, short SectorFlag=0);
   friend short Pmatrix::DoFix(short t, short lvn, short SectorFlag=0);
   friend char Vector::GetFixType(short Sectorno);
   friend short Vector::IsFixOpt(short sector, char flag);
};


#endif // end FIXBANK_H class definitions

