/* SEIDEL.CPP -- All definitions of Seidel() and PSeidel() have
	been moved here.
*/

#include "dymesys.h"
#include "user.h"     // for outfix

/* Seidel() Version 1, simplest form.****************************************

Solves by the Seidel process the equation
	q = Aq + f
Typically, q is output and f is final demand.  triang is the
triangularization vector as an IVector, and toler is the error tolerance in
testing convergence of the Seidel method.
*/
short Seidel(Matrix& A, Vector& q, Vector& f, IVector& triang,
	float toler)
{
	short i,j,k,first,last,n,iter,im1,imax;
	float discrep,dismax,fdismax;
	double sum;

	iter = 0; // Iteration count
	n = A.nr;
	first = A.fc;
	last =  A.lc;

	if(q.nelem < A.nr){
		printf("In Seidel, the solution vector is not large enough.\n");
		return(ERR);
		}
	while(1){
		dismax = 0;
		for(k = triang.First(); k < triang.First()+ triang.numelm(); k++){
			i = triang[k];
			im1 = i-first;
			sum = f.v[im1];
			for(j = first; j <= last; j++)
				sum += A.m[i][j]*q.v[j-1];
			sum -= A.m[i][i]*q.v[im1]; // Take off the diagonal element of sum
			sum = sum/(1.- A.m[i][i]);
			discrep = fabs(sum - q.v[im1]);
			if(discrep > dismax){
				dismax = discrep;
				imax = i;
				if(fabs(sum)>1.0e-7) {
					fdismax = fabs(sum);
					fdismax = dismax/fdismax;
					}
				}
			q.v[im1] = sum;
			arith("in Seidel",i);
			}
		iter++;
		if(dismax < toler || fdismax < .000001) break;
		if(iter > 25 ){
			printf("No convergence in %d iterations. Discrep = %7.2f in sector %d.\n",
				iter, dismax, imax);
			return(ERR);
			}
		}
	#if !defined(OS2COMPILE)
	// gotoxy(26,wherey());
	#endif
	printf("Seidel iterations: %d ",iter);
	return(OK);
	}

#ifdef INCL_IVECTOR

/* PSeidel() - For price equations, Version 1, simplest
This version uses the IVector class for triang
Solves by the Seidel process the equation
	p = pA + v
Typically, p is price; v, value added.  Itrian is the triangularization vector,
which will be used in reverse order.
*/
short PSeidel(Matrix& A, Vector& p,  Vector& v, IVector& triang, float toler)
{
	short i,j,k,first,last,n,iter,jm1,jmax;
	float discrep,dismax,fdismax;
	double sum;

	iter = 0; // Iteration count
	n = A.nc;
	first = A.fr;
	last =  A.lr;

	arith("before PSeidel",0);
	if(p.nelem < A.nc){
		printf("In PSeidel, the solution vector is not large enough.\n");
		return(ERR);
		}
	while(1){
		dismax = 0;
 	    for(k = triang.Last(); k >= triang.First(); k--){
		    j = triang[k];
			jm1 = j-first;
			sum = v.v[jm1];
			for(i = first; i <= last; i++)
				sum += A.m[i][j]*p.v[i-first];
			sum -= A.m[j][j]*p.v[jm1]; // Take off the diagonal element of sum
			sum = sum/(1.- A.m[j][j]);
			discrep = fabs(sum - p.v[jm1]);
			if(discrep > dismax){
				dismax = discrep;
				jmax = j;
				fdismax = fabs(sum);
				fdismax = dismax/fdismax;
				}
			p.v[jm1] = sum;
			arith("PSeidel",j);
			}
		iter++;
		if(dismax < toler || fdismax < .000001) break;
		if(iter > 25 ){
			printf("PSeidel does not converge in %d iterations.\nDiscrep = %7.2f in sector %d.\n",
				iter, dismax, jmax);
			return(ERR);
			}
		}
	return(OK);
	}
#endif


/* Seidel() Obsolete form of Version 1 **************************
Solves by the Seidel process the equation
	q = Aq + f
Typically, q is output; f, final demand;  triang is the
triangularization vector. Does not use IVector for triang.
Not recommended for new models.
*/
short Seidel(Matrix& A, Vector& q, Vector& f, short *triang,
	float toler)
{
	short i,j,k,first,last,n,iter,im1,imax;
	float discrep,dismax,fdismax;
	double sum;

	iter = 0; // Iteration count
	n = A.nr;
	first = A.fc;
	last =  A.lc;

	if(q.nelem < A.nr){
		printf("In Seidel, the solution vector is not large enough.\n");
		return(ERR);
		}
	while(1){
		dismax = 0;
		for(k = 0; k < n; k++){
			i = triang[k];
			im1 = i-first;
			sum = f.v[im1];
			for(j = first; j <= last; j++)
				sum += A.m[i][j]*q.v[j-1];
			sum -= A.m[i][i]*q.v[im1]; // Take off the diagonal element of sum
			sum = sum/(1.- A.m[i][i]);
			discrep = fabs(sum - q.v[im1]);
			if(discrep > dismax){
				dismax = discrep;
				imax = i;
				if(fabs(sum)>1.0e-7) {
					fdismax = fabs(sum);
					fdismax = dismax/fdismax;
					}
				}
			q.v[im1] = sum;
			arith("in Seidel",i);
			}
		iter++;
		if(dismax < toler || fdismax < .000001) break;
		if(iter > 25 ){
			printf("No convergence in %d iterations. Discrep = %7.2f in sector %d.\n",
				iter, dismax, imax);
			return(ERR);
			}
		}
	#if !defined(OS2COMPILE)
	//gotoxy(26,wherey());
	#endif
	printf("Seidel iterations: %d ",iter);
	return(OK);
	}


/* PSeidel() for price computations, Obsolete form of version 1 *************
PSolves by the Seidel process the equation
	p = pA + v
Typically, p is price; v, value added.  triang is the triangularization vector,
which will be used in reverse order.
Does not use IVector for triang; not recommended for new models.

*/
short PSeidel(Matrix& A, Vector& p,  Vector& v, short *triang, float toler)
{
	short i,j,k,first,last,n,iter,jm1,jmax;
	float discrep,dismax,fdismax;
	double sum;

	iter = 0; // Iteration count
	n = A.nc;
	first = A.fr;
	last =  A.lr;

	arith("before PSeidel",0);
	if(p.nelem < A.nc){
		printf("In PSeidel, the solution vector is not large enough.\n");
		return(ERR);
		}
	while(1){
		dismax = 0;
		for(k = 0; k < n; k++){
			j = triang[n-k-1];
			jm1 = j-first;
			sum = v.v[jm1];
			for(i = first; i <= last; i++)
				sum += A.m[i][j]*p.v[i-first];
			sum -= A.m[j][j]*p.v[jm1]; // Take off the diagonal element of sum
			sum = sum/(1.- A.m[j][j]);
			discrep = fabs(sum - p.v[jm1]);
			if(discrep > dismax){
				dismax = discrep;
				jmax = j;
				fdismax = fabs(sum);
				fdismax = dismax/fdismax;
				}
			p.v[jm1] = sum;
			arith("PSeidel",j);
			}
		iter++;
		if(dismax < toler || fdismax < .000001) break;
		if(iter > 25 ){
			printf("PSeidel does not converge in %d iterations.\nDiscrep = %7.2f in sector %d.\n",
				iter, dismax, jmax);
			return(ERR);
			}
		}
	#if !defined(OS2COMPILE)
	gotoxy(42,wherey());
	#endif
	cprintf("%d ",iter);
	return(OK);
	}
//**********************************************************

//===========================================================================
/* Seidel() packed matrix,  version 3 ***********************************
Solves by the Seidel process the equation
	q = Aq + D(m)q + f
where D(m) is a diagonal matrix with the elements of m on the diagonal.
Typically, q is output; f, final demand; m, the coefficients of the current
q in the import and inventory-change equations.  triang is the
triangularization vector.
*/
short Seidel(Pmatrix& A, Vector& q, Vector& f,  Vector& dump,	short *triang,
	float toler)
{
	short i,j,k,first,last,n,iter,im1,imax;
	float discrep,dismax,fdismax;
	double sum;
	short colno;

	iter = 0; // Iteration count
	n = A.nr;
	first = A.fc;
//	last =  A.lc;

	if(q.nelem < A.nr){
		printf("In Seidel, the solution vector is not large enough.\n");
		return(ERR);
		}
	while(1){
		dismax = 0;
		for(k = 0; k < n; k++){
			i = triang[k];
			im1 = i-first;
			sum = f.v[im1];
			if(A.rnc[i] > 0){
				for(j = 0; j < A.rnc[i]; j++){
					colno = A.row[i][j]; // the column number of the nozero coeff
					sum += A.p[i][j]*q.v[colno-1];
					}
				}
			sum -= A(i, i)*q.v[im1]; // Take off the diagonal element of sum
			if(outfix[i] == 'e') //Use the equation
				sum = sum/(1.- A(i, i));
			else{		  // Dump error into dump
				dump[i] = q.v[im1] -(sum + A(i,i)*q.v[im1]);
				sum = q.v[im1];
				}
			discrep = fabs(sum - q.v[im1]);
			if(discrep > dismax){
				dismax = discrep;
				imax = i;
				if(fabs(sum)>1.0e-7) {
					fdismax = fabs(sum);
					fdismax = dismax/fdismax;
					}
				}
			q.v[im1] = sum;
			arith("in Seidel",i);
			}
		iter++;
		if(dismax < toler || fdismax < .000001) break;
		if(iter > 25 ){
			printf("No convergence in %d iterations. Discrep = %7.2f in sector %d.\n",
				iter, dismax, imax);
			return(ERR);
			}
		}
	#if !defined(OS2COMPILE)
	gotoxy(26,wherey());
	#endif
	cprintf("%d ",iter);
	return(OK);
	}

/* Seidel() with packed matrix ****************************************
Solves by the Seidel process the equation
	q = Aq + D(m)q + f
where D(m) is a diagonal matrix with the elements of m on the diagonal.
Typically, q is output; f, final demand; m, the coefficients of the current
q in the import and inventory-change equations.  triang is the
triangularization vector.

This version uses a simple import equation with import fixes, including
group fixes.
*/
/*
short Seidel(Pmatrix& A, Vector& q, Equation& ImportEq, Vector& fd,
       Vector& im, Vector& ven, Vector& ex, Vector& def,
		 Vector& dump, short *triang, float toler)
{
	short i,j,k,first,last,n,iter,im1,imax,ieq;
	float discrep,dismax,fdismax;
   float hithere;
	float dd;
	float coef2;
	double sum;
	short colno,iterSave;
	short ImportFreeze=FALSE;	// Set this to skip import calculation.

	n = A.nr;
	first = A.fc;
//	last =  A.lc;

	if(q.nelem < A.nr){
		printf("In Seidel, the solution vector is not large enough.\n");
		return(ERR);
		}

StartSeidel:
	iter = 0; // Iteration count
	while(1){
		dismax = 0;
		for(k = 0; k < n; k++){
			i = triang[k];
			im1 = i-first;
			sum = fd.v[im1];
			sum += ven[i];  // Eventually, inventory change eqs would go here.

			// Add intermediate
			for(j = 0; j < A.rnc[i]; j++){
				colno = A.row[i][j];
				sum += A.p[i][j]*q.v[colno-1];
				}
			sum -= A(i, i)*q.v[im1]; // Take off the diagonal element of sum

			if(t!=godate && !ImportFreeze) {
				// don't try to calculate both discrepancy and imports in the
					// first year.
				// calculate imports equation:
				ieq = ImportEq.GetLoc(i);
				if(ieq==-1) im[i]=0.0;
				else  {
					if(ImportEq.type(ieq)=='l') // (leave data in file)
						;
					// Simple linear regression on domestic demand:
					else {
						dd = sum - ex[i] - def[i];  // Margaret's suggestion for dd
						im[i] = ImportEq[ieq][1] * dd;
						}
					}
				}
			sum -= im[i];

			if(outfix[i] == 'e') {
				coef2=A(i,i);  // debugging
				sum = sum/(1.-coef2);

//				sum = (sum -im[i])
 //						/((1.- A(i, i))*(1-im[i])); // use actual im, til we get eq.
 				}
			else {
				sum += A(i,i)*q.v[im1];
				dump[i] = q.v[im1] -(sum - im[i]);
				sum = q.v[im1];
				}
			discrep = fabs(sum - q.v[im1]);
			if(discrep > dismax){
				dismax = discrep;
				imax = i;
				fdismax = fabs(sum);
				fdismax = dismax/fdismax;
				}
			q.v[im1] = sum;
			arith("Seidel",i);
			}
		iter++;
		if(dismax < toler || fdismax < .000001) break;
		if(iter > 25 ){
			printf("No convergence in %d iterations. Discrep = %7.2f in sector %d.\n",
				iter, dismax, imax);
			return(ERR);
			}
		}


   for(i=1;i<n;i++)
      fd[i] += (ven[i] - im[i]);
	#if !defined(OS2COMPILE)
	gotoxy(26,wherey());
	#endif
	cprintf("%d",iterSave);
	return(OK);
	}
*/
