// Model5.cpp
extern char *Version;    // Interdyme Version
char *SlimDymeDate="April 2006";
char *ModelTitle="TINY - Pedagogic Interindustry Dynamic Model [26x26]";
/*
  MODEL.CPP -- This file consists of the loop function, plus any
  equation functions that comprise the part of the model written
  by the Interdyme user.

  This file, and the CONFIG.CPP file, should be the only two .CPP
  files that the model builder needs to work on.
*/
#include "dymesys.h"  // All includes for general Interdyme "infrastructure".
#if defined(INCL_MACRO)
#include "tseries.inc" // Re-written each time by IDBUILD.
#include "heart.h"
#endif
#include "user.h"     // All global variables for user model.

// Proto type for Seidel used in TINY
short Seidel(Matrix& A, Vector& q, Vector& f, IVector& triang,
	float toler);

/* Prototypes for other versions of Seidel
short Seidel(Matrix& A, Vector& q, Vector& imp, Vector& sdc,
        Equation& imports, Vector& f, short *triang,    float toler);
short PSeidel(Matrix& A, Vector& p, Vector& v, short *triang,
        float toler);
// With import share matrix:
short Seidel(Matrix& A, Matrix& MS, Vector& q, Vector& f, short *triang,
        float toler);
short PSeidel(Matrix& A, Matrix& MS, Vector& p, Vector& v, short *triang,
        float toler);

short Seidel(Matrix& A, Vector& q, Vector& imp, Vector& sdc,
        Equation& imports, Vector& f, IVector& triang, float toler);
short PSeidel(Matrix& A, Vector& p,  Vector& v, IVector& triang, float toler);

*/


/* ====================================================================*/
/* loop()is the start-up routine of the forecasting model.  It resizes the
   matrices and vectors of the VAM file which which are declared globally
   in user.h.  Similarly, resizes the equation objects that have been
   estimated in G and globally declared in user.h. Finally, it either
   calls spin() to cycle through the years, calling the various functions
   that do the forecasting, or it calls maxsolve to optimize through
   many calls to spin().
*/

const int NSEC=8; // Number of sectors in I-O table, most vectors.

void spin();
int pcefunc();
int importfunc(Vector& q);
short Seidel(Matrix& A, Vector& q, Vector& fd, Vector& dump, IVector& triang,
	float toler);

void loop(void)
{

    /* Resizing of vectors and matrices from the VAM file is done here.
       These vectors and matrices are declared globally in user.h, so they
       can be accessed by any function anywhere in the program.
       This declaration, however, cannot allocate space for them.  Essentially,
       they all have 0 rows and 0 columns.  Giving them real sizes and
       allocating space for them is done here by calls to each object's
       "resize" function, denoted by just ".r".

     The format is:
    <Dymename>.r("<vamname>",<readflag>,<writeflag>)

        where <Dymename> is what the object will be called in this program.
                <vamname>  is what it is called in VAM.
                <readflag> is 'i' if item is to be read only in the initial year.
                            'n' if it is NEVER to be read.
                            'y' otherwise.
                <writeflag> is 'n' if the item is not to be written.
    Normally, <Dymename> is the same as <vamname>.
    Default values of readflag = 'y' and writeflag = 'y' are supplied
    if they are omitted in the declaration.  Vectors with lagged values
    should be declared both as a Vector and as a Matrix, with the
    Vector definition coming first.

    Interdyme provides for special treatment of large, sparse matrices.
    The non-zero elements of the matrices may be store in a special form
    called "packed matrices."  The packed matrices do not go into the VAM
    file; rather, each has its own file.  Here is an example.
       Example of Packed Matrix (Pmatrix):
       Pmatrix B("bmv",'a','y', BmvFileName);

    Example of vectors with lagged values.  The resizing as a Matrix
    cannot precede the resizing as a Vector.  Do not use the same name
    or one differing only by capitalization for both the Vector and Matrix.

     Outlag.r("q",'i','n'), Priceslag.r("pq",'i','n');

    Second vam file:
    If you were reading some vectors out of a second Vam file,
    Vector declarations could be as follows:

    Vector fd(vf[1],"fd",'y','n'), va(vf[1],"va",'y','n');
        or:
    VamFile& vfScratch=vf[2];
    Vector fd(vfScratch,"fd",'y','n'), va(vfScratch,"va",'y','n');
    */

   // Vectors
   out.r("out"); pce.r("pce"); gov.r("gov"); 
   inv.r("inv"); ex.r("ex"); im.r("im"); fd.r("fd"); 
   outN.r("outN"); pceN.r("pceN"); govN.r("govN"); invN.r("invN"); 
   exN.r("exN"); imN.r("imN"); fdN.r("fdN"); dep.r("dep"); 
   lab.r("lab"); cap.r("cap"); ind.r("ind"); va.r("va"); 
   depc.r("depc"); labc.r("labc"); capc.r("capc"); indc.r("indc"); 
   vac.r("vac"); pcec.r("pcec"); invc.r("invc"); govc.r("govc"); 
   exc.r("exc"); imc.r("imc"); prices.r("prices"); x.r("x"); 
   y.r("y"); fix.r("fix");

   // Matrices
    FM.r("FM"); AM.r("AM");

    #ifdef INCL_IVECTOR
    // The triang vector specifies the order of solving equations in Seidel.
    // Example of how to use IVector class for all triangulation.
    triang.resize(NSEC);
    triang.ReadA("triang.iv");   // Read the triangularization vector
    #endif

   lastdata(); // Use of lastdata function requires that the LastData file exist.

    // Resize and read the vecor for spreading the PCE discrepancy.
    PCESpread.resize(NSEC);
    PCESpread.ReadA("PCESpread.dat");

    // Resize, read, and store the PCE equations; reads the pce.eqn file.
    pceeqn.r("pce.eqn");
    impeqn.r("import.eqn");

    // pceeqn.rhostart is read as the last year used in estimating the PCE equations.
    // We want it to be the year in which to start the rhoadjustment.  If we are
    // doing a historical simulation, that should MacEqStartDate if it is before the
    // date recorded in the pce.eqn file.
    if(pceeqn.rhostart > MacEqStartDate) pceeqn.rhostart = MacEqStartDate;

    if(MaxFlag == 'n')
    	spin();
    else{
    	maxsolve();
    	MaxFlag = 'n';
    	spin();
    	}
   #if defined(INCL_MACRO)
       storets(); // Store the macro variable time series
   #endif

   printf("\nThe run has finished.  Use G or Compare to view results.\n");
   tap();
   }

// The spin() function sends the model through one cycle of years.

void spin(){
   // General start of the spin() function:
    short Iteration,Seiditer;
    #if defined(INCL_MACRO)
    tserin(); //load macrovariables
    #endif
    // Set tolerance for Seidel process
    toler = 1.;
    float oldpcetot,oldinvtot,pcedif,invdif;
    float pcesum, pcediscrep;
	setrho = 'n';
	
    Vector dump(NSEC);
    // Iteration loop period-by-period:
    // ===============================
    if (MaxFlag == 'n') printf("\n");
    for (t = godate; t<= stopdate; t++) {
       if (MaxFlag == 'n') printf("%d ",t);
       // Load all vectors and matrices.
       load(t);
	uptseries();
	Iteration = 0;
	// Start of code particular to TINY:
	vac = depc+labc+capc+indc;
	PSeidel(AM, prices, vac, triang, 0.000001);
	// The loop for convergence on pcetot and invtot.
       while(Iteration < 20){
    	    Iteration++;
	    oldinvtot = invtot[t]; oldpcetot = pcetot[t];
	    // Compute the estimated PCE equations with equation system
	    if(t>= MacEqStartDate){
	        pcefunc();
	        // Sum up the calculated PCE elements
 	        pcesum = pce.sum();
	        pcediscrep = pcetot[t] - pcesum;
	        // printf("\npcediscrep = %10.2f\n",pcediscrep);
	        // Spread the discrepancy by the proportions of the PCESpread vector.
	        pce = pce + pcediscrep*PCESpread;
	        invtotf();
	        }
	    inv = invtot[t]*invc;
	    gov = govtot[t]*govc;
	    ex  = extot[t]*exc;
	    fd = pce + gov + inv + ex;
           Seiditer = Seidel(AM, out, fd, dump, triang, toler);
           imtot[t] = im.sum();
	    dep = ebemul(depc,out);  lab = ebemul(labc,out);
    	    cap = ebemul(capc,out);  ind = ebemul(indc,out);

	    // The Accountant for Tiny
	    gdp[t] = fd.sum()- imtot[t];
	    if(t>MacEqStartDate){
	        deprec[t] = dep.sum(); labinc[t] = lab.sum();
	        capinc[t] = cap.sum(); indtax[t] = ind.sum();
               accountf();
	        }
           invdif = fabs(invtot[t] - oldinvtot);
	    pcedif = fabs(pcetot[t]- oldpcetot);
	    printf("Spin Iter %2d SeidIter %2d pce = %7.1f pcedif = %6.2f invdif = %6.2f\n",
			Iteration,Seiditer,pcetot[t],pcedif,invdif);
	    if(invdif < .5 && pcedif < .5) break;
	    }
	// Set error for rhoadjustment
	setrho = 'y';
	if(t >= pceeqn.rhostart)  pcefunc();
       if(t >= impeqn.rhostart)  importfunc(out);
	setrho = 'n';
	// Here when both Investment and PCE have converged
	// Compute flow matrix in current prices
       FM = AM % out;
	FM = prices % FM;
       // Put output and final demands into nominal prices
       outN = ebemul(prices,out);
       fdN = ebemul(prices,fd);
       pceN = ebemul(prices,pce);
       invN = ebemul(prices,inv);
       exN  = ebemul(prices,ex);
       imN  = ebemul(prices,im);
       govN = ebemul(prices,gov);
	// End of code particular to TINY
	// Standard end of the spin() function:
	if(MaxFlag == 'y')
           shiftback(t);
       else{
          // Store the values of vectors and matrices for this period.
          store(t);
          printf("\n");
          }
	}
   }

// pcefunc() -- PCE functions for TINY
int pcefunc(){
	int n, i,j,t1;
	float cons;
	char which;

	n = pceeqn.neq; // Number of equations
	// pdisinc is personal disposable income. It is a global macro variable.
	// Compute variables used in several equations.
	dpdis[t] = pdisinc[t] - pdisinc[t-1];

	// Loop over the equations
	for(i = 1; i <= n; i++){
		j = pceeqn.sec(i); // j is the sector number of this equation.
		which = pceeqn.type(i);
		if(which == 'L'){
			cons = pceeqn[i][1] + pceeqn[i][2]*pdisinc[t] + pceeqn[i][3]*dpdis[t];
			}
		else{
			printf("Unknown equation type %c in pcefunc, category %d.\n",
				which,i);
			tap(); //Pause so that error message can be read.
			continue;
			}
		// Apply rhoadjustment			}
		// Note the use of i and j in the following statement.
	    pce[j] = pceeqn.rhoadj(cons,pce[j],i);
		}
	pce.fix(t);
	return(n);
	}
// Importfunc() -- Import functions for TINY
int importfunc(Vector& q){
	int n, i,j,t1;
	float imp;
	char which;

    if(t < MacEqStartDate) return(0);
	n = impeqn.neq; // Number of equations
	// pdisinc is personal disposable income. It is a global macro variable.

	// Loop over the equations
	for(i = 1; i <= n; i++){
    	j = impeqn.sec(i); // j is the sector number of this equation.
		which = impeqn.type(i);
		if(which == 'L'){
			imp = impeqn[i][1] + impeqn[i][2]*q[j];
			}
		else{
			printf("Unknown equation type %c in importfunc, sector %d.\n",
				which,i);
			tap(); //Pause so that error message can be read.
			continue;
			}
		// Apply rhoadjustment			}
		// Note the use of i and j in the following statement.
	    im[j] = impeqn.rhoadj(imp,im[j],i);
		}
	im.fix(t);
	return(n);
	}
// Seidel() Version 2, with import equations**

short Seidel(Matrix& A, Vector& q, Vector& fd, Vector& dump, IVector& triang,
	float toler)
{
	short i,j,k,first,last,n,iter,im1,imax;
	float discrep,dismax,fdismax;
	double sum;
	Vector f(fd.rows()); //scratch vector

	iter = 0; // Iteration count
	n = A.rows();
	first = A.firstcolumn();
	last =  A.lastcolumn();

	if(q.rows() < A.rows()){
		printf("In Seidel, the solution vector is not large enough.\n");
		return(ERR);
		}
	dump.set(0.);
	while(1){
           // Compute imports
           importfunc(q);
	    // Reduce final demand by imports
           f = fd - im;
	    dismax = 0;
	    for(k = triang.First(); k < triang.First()+ triang.numelm(); k++){
		i = triang[k];
		sum = f[i];
		for(j = first; j <= last; j++)
			sum += A[i][j]*q[j];
		sum -= A[i][i]*q[i]; // Take off the diagonal element of sum
		sum = sum/(1.- A[i][i]);
		if(outfix[i] != 'e'){ // Override the equation
		    // Dump excess into dump
		    dump[i] = q[i] - sum;
		    sum = q[i];
		    }
		discrep = fabs(sum - q[i]);
		if(discrep > dismax){
		    dismax = discrep;
		    imax = i;
		    if(fabs(sum)>1.0e-3) {
			fdismax = fabs(sum);
			fdismax = dismax/fdismax;
			}
		    }
		q[i] = sum;
		arith("in Seidel",i);
		}
	   iter++;
	   if(dismax < toler || fdismax < .000001) break;
	   if(iter > 25 ){
		printf("No convergence in %d iterations. Discrep = %7.2f in sector %d.\n",
			iter, dismax, imax);
		return(ERR);
		}
	   }
	//printf("Seidel iterations: %d ",iter);
	return(iter);
	}
